import path from "node:path";
import { Glob } from "bun";
import z from "zod";
import {
  getEmptyPlaybill,
  type AnyResource,
  type Playbill,
} from "@proscenium/playbill";
import { loadYAMLFileOrFail } from "./files";
import { loadResourceFile } from "./resource";

const BindingFileSchema = z.object({
  $binding: z.literal("playbill"),
  extends: z.string().optional(),
  id: z.string(),
  name: z.string().default("Unnamed playbill"),
  author: z.string().optional(),
  description: z.string().optional(),
  version: z.string(),
  files: z.array(z.string()).default(["**/*.yaml", "**/*.md"]),
});

type BindingFileContent = z.infer<typeof BindingFileSchema>;

export interface PlaybillBinding {
  _raw: BindingFileContent;

  // File information
  bindingFilePath: string;
  includedFiles: string[];

  // Binding properties
  id: string;
  name: string;
  author: string;
  version: string;
  description: string;

  playbill: Playbill;
}

export async function loadFromBinding(
  filePath: string,
): Promise<PlaybillBinding> {
  const yamlContent = await loadYAMLFileOrFail(filePath);
  const binding = BindingFileSchema.parse(yamlContent);
  const fileGlobs = binding.files;
  const bindingFileDirname = filePath.split("/").slice(0, -1).join("/");

  let basePlaybill: Playbill | null = null;

  // If this is extending another binding, load that first
  if (binding.extends) {
    console.log(binding);
    const extendBindingPath = path.resolve(bindingFileDirname, binding.extends);
    basePlaybill = (await loadFromBinding(extendBindingPath)).playbill;
  }

  const allFilePaths: string[] = [];

  for (const thisGlob of fileGlobs) {
    const glob = new Glob(thisGlob);

    const results = glob.scanSync({
      cwd: bindingFileDirname,
      absolute: true,
      followSymlinks: true,
      onlyFiles: true,
    });

    allFilePaths.push(...results);
  }

  const bindingFileAbsolutePath = path.resolve(filePath);
  const filePathsWithoutBindingFile = allFilePaths.filter((filePath) => {
    return filePath !== bindingFileAbsolutePath;
  });

  // -- LOAD ASSOCIATED RESOURCE FILES -- //
  const loadedResources: AnyResource[] = [];
  for (const filepath of filePathsWithoutBindingFile) {
    const loaded = await loadResourceFile(filepath);
    loadedResources.push(...loaded);
  }

  // -- COMPILE PLAYBILL FROM RESOURCES --//
  const playbill = basePlaybill ?? getEmptyPlaybill();

  playbill.id = binding.id;
  playbill.name = binding.name;
  playbill.author = binding.author ?? "Anonymous";
  playbill.version = binding.version;
  playbill.description = binding.description ?? "";

  for (const resource of loadedResources) {
    switch (resource.$define) {
      case "ability":
        playbill.abilities.push(resource);
        break;
      case "species":
        playbill.species.push(resource);
        break;
      case "entity":
        playbill.entities.push(resource);
        break;
      case "item":
        playbill.items.push(resource);
        break;
      case "tag":
        playbill.tags.push(resource);
        break;
      case "method":
        playbill.methods.push(resource);
        break;
      case "lore":
        playbill.lore.push(resource);
        break;
      case "rule":
        playbill.rules.push(resource);
        break;
      case "playbill":
        throw new Error("Cannot load playbills rn dawg");
    }
  }

  return {
    _raw: binding,
    bindingFilePath: filePath,
    includedFiles: filePathsWithoutBindingFile,

    id: binding.id,
    name: binding.name,
    author: binding.author ?? "Anonymous",
    description: binding.description ?? "",
    version: binding.version,

    playbill,
  };
}
