import {
  type AnyResource,
  AnyResourceSchema,
  ResourceMap,
} from "@proscenium/playbill";
import YAML, { YAMLParseError } from "yaml";
import { ZodError } from "zod";
import { MalformedResourceFileError } from "./errors";
import { extractFrontmatter, extractMarkdown } from "./markdown";
import { loadFileOrFail } from "./files";

type FileFormat = "yaml" | "markdown";

function determineFileFormat(filePath: string): FileFormat {
  return filePath.split(".").pop() === "md" ? "markdown" : "yaml";
}

function loadYamlResourceFile(filePath: string, text: string): AnyResource[] {
  const parsedDocs = YAML.parseAllDocuments(text);

  if (parsedDocs.some((doc) => doc.toJS() === null)) {
    throw new MalformedResourceFileError("Encountered NULL resource", filePath);
  }

  const errors = parsedDocs.flatMap((doc) => doc.errors);

  if (errors.length > 0) {
    throw new MalformedResourceFileError(
      "Error parsing YAML resource",
      filePath,
      errors.map((e) => e.message).join(", "),
    );
  }

  const collection: AnyResource[] = [];

  for (const doc of parsedDocs) {
    const raw = doc.toJS();
    const parsed = AnyResourceSchema.safeParse(raw);

    if (!parsed.success) {
      throw new MalformedResourceFileError(
        "Failed to parse resource",
        filePath,
        `Schema validation failed: ${parsed.error.message}`,
      );
    }

    const parsedResource = parsed.data;
    const type = parsedResource.$define;
    const schemaToUse = ResourceMap[type];
    const validated = schemaToUse.parse(parsedResource);
    collection.push(validated);
  }

  return collection;
}

function loadMarkdownResourceFile(
  filePath: string,
  text: string,
): AnyResource[] {
  try {
    const frontmatter = extractFrontmatter(text);
    const markdown = extractMarkdown(text);

    const together = {
      ...frontmatter,
      description: markdown,
    };

    const parsed = AnyResourceSchema.parse(together);
    return [parsed];
  } catch (e) {
    if (e instanceof YAMLParseError) {
      throw new MalformedResourceFileError(
        "Error parsing Markdown frontmatter",
        filePath,
        e.message,
      );
    }

    if (e instanceof ZodError) {
      throw new MalformedResourceFileError(
        "Error parsing resource file",
        filePath,
        e.message,
      );
    }

    throw e;
  }
}

/**
 * Load and process all documents in a yaml file at the given path
 * @param filePath - The path to the yaml file
 * @returns An array of validated resources
 *
 * @throws ResourceFileError
 * @throws NullResourceError
 */
export async function loadResourceFile(
  filePath: string,
): Promise<AnyResource[]> {
  const text = await loadFileOrFail(filePath);
  const format = determineFileFormat(filePath);

  switch (format) {
    case "yaml":
      return loadYamlResourceFile(filePath, text);
    case "markdown":
      return loadMarkdownResourceFile(filePath, text);
  }
}
